@@ -21,22 +21,24 @@
 #include <QPair>
 
 #include "qgis_core.h"
-#include "qgsfields.h"
 #include "qgsreadwritecontext.h"
+#include "qgsrelationcontext.h"
+#include "qgsattributes.h"
 
 #include "qgis_sip.h"
 
 class QgsFeatureIterator;
 class QgsFeature;
 class QgsFeatureRequest;
-class QgsAttributes;
 class QgsVectorLayer;
 class QgsRelationPrivate;
-
+class QgsPolymorphicRelation;
 
 /**
  * \ingroup core
  * \class QgsRelation
+ *
+ * \brief Represents a relationship between two vector layers.
  */
 class CORE_EXPORT QgsRelation
 {
@@ -47,25 +49,17 @@ class CORE_EXPORT QgsRelation
     Q_PROPERTY( QgsVectorLayer *referencedLayer READ referencedLayer )
     Q_PROPERTY( QString name READ name WRITE setName )
     Q_PROPERTY( bool isValid READ isValid )
+    Q_PROPERTY( QString polymorphicRelationId READ polymorphicRelationId WRITE setPolymorphicRelationId )
+    Q_PROPERTY( QgsPolymorphicRelation polymorphicRelation READ polymorphicRelation )
 
   public:
 
-    /**
-     * enum for the relation strength
-     * Association, Composition
-     */
-    enum RelationStrength
-    {
-      Association, //!< Loose relation, related elements are not part of the parent and a parent copy will not copy any children.
-      Composition  //!< Fix relation, related elements are part of the parent and a parent copy will copy any children or delete of parent will delete children
-
-    };
-
 #ifndef SIP_RUN
 
     /**
      * \ingroup core
-     * Defines a relation between matching fields of the two involved tables of a relation.
+     * \brief Defines a relation between matching fields of the two involved tables of a relation.
+     *
      * Often, a relation is only defined by just one FieldPair with the name of the foreign key
      * column of the referencing (child) table as first element and the name of the primary key column
      * of the referenced (parent) table as the second element.
@@ -96,6 +90,11 @@ class CORE_EXPORT QgsRelation
     QgsRelation();
     ~QgsRelation();
 
+    /**
+     * Constructor with context. Creates an invalid relation.
+     */
+    QgsRelation( const QgsRelationContext &context );
+
     /**
      * Copies a relation.
      * This makes a shallow copy, relations are implicitly shared and only duplicated when the copy is
@@ -115,10 +114,11 @@ class CORE_EXPORT QgsRelation
      *
      * \param node The dom node containing the relation information
      * \param context to pass project translator
+     * \param relationContext a relation context
      *
      * \returns A relation
      */
-    static QgsRelation createFromXml( const QDomNode &node, QgsReadWriteContext &context );
+    static QgsRelation createFromXml( const QDomNode &node, QgsReadWriteContext &context, const QgsRelationContext &relationContext = QgsRelationContext() );
 
     /**
      * Writes a relation to an XML structure. Used for saving .qgs projects
@@ -140,9 +140,8 @@ class CORE_EXPORT QgsRelation
 
     /**
      * Set a strength for this relation
-     * \since QGIS 3.0
      */
-    void setStrength( RelationStrength strength );
+    void setStrength( Qgis::RelationshipStrength strength );
 
     /**
      * Set the referencing (child) layer id. This layer will be searched in the registry.
@@ -155,7 +154,7 @@ class CORE_EXPORT QgsRelation
     void setReferencedLayer( const QString &id );
 
     /**
-     * Add a field pairs which is part of this relation
+     * Add a field pair which is part of this relation
      * The first element of each pair are the field names of the foreign key.
      * The second element of each pair are the field names of the matching primary key.
      *
@@ -165,7 +164,7 @@ class CORE_EXPORT QgsRelation
     void addFieldPair( const QString &referencingField, const QString &referencedField );
 
     /**
-     * Add a field pairs which is part of this relation
+     * Add a field pair which is part of this relation
      * The first element of each pair are the field names of the foreign key.
      * The second element of each pair are the field names of the matching primary key.
      *
@@ -205,7 +204,6 @@ class CORE_EXPORT QgsRelation
      * \returns expression filter string for all the referencing features
      * \see getRelatedFeatures()
      * \see getRelatedFeaturesRequest()
-     * \since QGIS 2.16
      */
     QString getRelatedFeaturesFilter( const QgsFeature &feature ) const;
 
@@ -235,7 +233,8 @@ class CORE_EXPORT QgsRelation
      *
      * \param feature A feature from the referencing (child) layer
      *
-     * \returns A request the referenced feature
+     * \returns The referenced (parent) feature, or an invalid feature if no matching feature
+     * was found
      */
     QgsFeature getReferencedFeature( const QgsFeature &feature ) const;
 
@@ -252,9 +251,8 @@ class CORE_EXPORT QgsRelation
      * Returns the relation strength as a string
      *
      * \returns strength
-     * \since QGIS 3.0
      */
-    RelationStrength strength() const;
+    Qgis::RelationshipStrength strength() const;
 
     /**
      * A (project-wide) unique id for this relation
@@ -265,7 +263,6 @@ class CORE_EXPORT QgsRelation
 
     /**
      * Generate a (project-wide) unique id for this relation
-     * \since QGIS 3.0
      */
     void generateId();
 
@@ -336,34 +333,49 @@ class CORE_EXPORT QgsRelation
      */
     QgsAttributeList referencingFields() const;
 
+    /**
+     * Returns TRUE if none of the referencing fields has a NOT NULL constraint.
+     * \since QGIS 3.28
+     */
+    bool referencingFieldsAllowNull() const;
+
     /**
      * Returns the validity of this relation. Don't use the information if it's not valid.
      * A relation is considered valid if both referenced and referencig layers are valid.
      *
-     * \returns true if the relation is valid
+     * \returns TRUE if the relation is valid
+     *
+     * \see validationError()
      */
     bool isValid() const;
 
+    /**
+     * Returns a user-friendly explanation for why the relationship is invalid.
+     *
+     * Returns an empty string if the relationship isValid().
+     *
+     * \see isValid()
+     * \since QGIS 3.28
+     */
+    QString validationError() const;
+
     /**
      * Compares the two QgsRelation, ignoring the name and the ID.
      *
      * \param other The other relation
-     * \returns true if they are similar
-     * \since QGIS 3.0
+     * \returns TRUE if they are similar
      */
     bool hasEqualDefinition( const QgsRelation &other ) const;
 
     /**
      * Gets the referenced field counterpart given a referencing field.
      *
-     * \since QGIS 3.0
      */
     Q_INVOKABLE QString resolveReferencedField( const QString &referencingField ) const;
 
     /**
      * Gets the referencing field counterpart given a referenced field.
      *
-     * \since QGIS 3.0
      */
     Q_INVOKABLE QString resolveReferencingField( const QString &referencedField ) const;
 
@@ -375,13 +387,52 @@ class CORE_EXPORT QgsRelation
      */
     void updateRelationStatus();
 
+    /**
+     * Sets the parent polymorphic relation id.
+     * \since QGIS 3.18
+     */
+    void setPolymorphicRelationId( const QString &polymorphicRelationId );
+
+    /**
+     * Returns the parent polymorphic relation id. If the relation is a normal relation, a null string is returned.
+     * \since QGIS 3.18
+     */
+    QString polymorphicRelationId() const;
+
+    /**
+     * Returns the parent polymorphic relation. If the relation is a normal relation, an invalid polymorphic relation is returned.
+     * \since QGIS 3.18
+     */
+    QgsPolymorphicRelation polymorphicRelation() const;
+
+    /**
+     * Returns the type of the relation
+     * \since QGIS 3.18
+     */
+    Qgis::RelationshipType type() const;
+
+    /**
+     * Returns a user-friendly translated string representing a relationship \a cardinality.
+     *
+     * \since QGIS 3.28
+     */
+    static QString cardinalityToDisplayString( Qgis::RelationshipCardinality cardinality );
+
+    /**
+     * Returns a user-friendly translated string representing a relationship \a strength.
+     *
+     * \since QGIS 3.28
+     */
+    static QString strengthToDisplayString( Qgis::RelationshipStrength strength );
+
   private:
 
     mutable QExplicitlySharedDataPointer<QgsRelationPrivate> d;
+
+    QgsRelationContext mContext;
 };
 
 // Register QgsRelation for usage with QVariant
 Q_DECLARE_METATYPE( QgsRelation )
-Q_DECLARE_METATYPE( QgsRelation::RelationStrength )
 
 #endif // QGSRELATION_H
